`timescale 1ns/1ps

module DataCache(  
	input 			clk,
	input 			reset,
	
	//Connections to CoherencyEngine
	input	[31:0]	dc_DataFromMem,
	output	[31:0]	dc_DataToMem,
	output	[29:0]	dc_RequestAddr,
	output			dc_RequestRead,
	output			dc_RequestWrite,
	input			dc_RequestReady,
	
	//Connections to Processor
	output	[31:0]	dc_DataToProcessor,
	input	[31:0]	dc_DataFromProcessor,
	input	[31:0]	dc_Address,
	output			dc_Stall,
	input			dc_Read,
	input			dc_Write,
	input	[1:0]	dc_Size
	);
	
	localparam stIDLE 				=0;
	
	localparam stWriteProc0			=5;
	localparam stWriteProc1        	=6;
	localparam stWriteProc2	   		=7;
	localparam stWriteProc3			=8;
	
	localparam stReadProc0			=9;
	localparam stReadProc1			=10;
	localparam stReadProc2			=11;
	localparam stReadProc3			=12;
	
	localparam BYTE =2'b01;
	localparam HALF =2'b10;
	localparam WORD =2'b11;
	
	
	//32 Bit Address:
	//Bits:     Use:
	//1-0		Byte Address
	//3-2		Word Address
	//(LINENUMBITS+3)-4		Line
	//31-(LINENUMBITS+4)	Tag
	
	//Parameter Definitions
	parameter NUMLINES = 2;
	parameter LINENUMBITS = 4;
	parameter TAGBITS = 32 - LINENUMBITS - 4;
	
	localparam ValidBit =0;
	localparam DirtyBit =1;
	
	//cacheLine declarations
	reg [127:0] cacheLineData [0:NUMLINES-1];
	reg [TAGBITS-1:0] cacheLineTag [0:NUMLINES-1];
	reg [1:0] cacheLineStatusBits [0:NUMLINES-1];
	
	//wires to hold sections of addresses
	wire [LINENUMBITS-1:0] ProcLineNum;
	wire [TAGBITS-1:0] ProcReqTag;
	
	//get line number and tag bits from processor address
	assign ProcLineNum = dc_Address[(LINENUMBITS+3):4];
	assign ProcReqTag = dc_Address[31:LINENUMBITS+4];
	
	//wires containing line addressed by LINENUM portion of address
	wire [127:0] ProcLine;
	wire [TAGBITS-1:0] ProcTag;
	wire ProcLineValid, ProcLineDirty;
	wire ProcTagMatch;
	
	assign ProcLine = cacheLineData[ProcLineNum];
	assign ProcTag = cacheLineTag[ProcLineNum];
	assign ProcLineValid = cacheLineStatusBits[ProcLineNum][ValidBit];
	assign ProcLineDirty = cacheLineStatusBits[ProcLineNum][DirtyBit];
	assign ProcTagMatch = ProcTag == ProcReqTag;
	
	
	assign dc_Stall = (dc_Read || dc_Write) 
					&& (!ProcLineValid 
						|| !ProcTagMatch);
	
	reg [29:0]  RequestAddress;	 
	reg [31:0]  RequestWriteData;
	reg			RequestRead, RequestWrite;	
	reg [4:0] State;				
	//if the state machine is idle, use the current address from 
	//the processor as the request address to the memory controller
	assign dc_RequestAddr = (State == stIDLE) ? dc_Address[31:4] : RequestAddress;
	
	//Invalidation is done when InvalidateLineValid is false, or the tags do not match
	assign dc_DataToMem = RequestWriteData;
						
	assign dc_RequestRead = RequestRead;
	assign dc_RequestWrite = RequestWrite;
	
	
	//Assign output to the processor based on cache line, etc.	
	wire [31:0] selectedWord;  
	wire [31:0] dc_DataToProcessor;
	
	assign selectedWord = 
	(dc_Address[3:2] == 2'b00) ? cacheLineData[ProcLineNum][31:0] :
	(dc_Address[3:2] == 2'b01) ? cacheLineData[ProcLineNum][63:32] :
	(dc_Address[3:2] == 2'b10) ? cacheLineData[ProcLineNum][95:64] :
								 cacheLineData[ProcLineNum][127:96];
	
	assign dc_DataToProcessor = selectedWord;
	
	
	always @(posedge clk) begin
		if(reset) begin
			State <= stIDLE;
			RequestRead <= 1'b0;
			RequestAddress <= 0;
			RequestWrite <= 1'b0;
		end
		else begin
	   	case(State)
			stIDLE: begin
				if (~ProcTagMatch && ProcLineValid && ProcLineDirty && (dc_Read || dc_Write)) begin
						$display("%d: DCache: Tag Mismatch on Dirty Line, Writing Back - ReqTag: %h, LineTag: %h",$time,ProcReqTag,ProcTag);
						RequestAddress <=  {ProcTag, ProcLineNum,2'b00}; //set address to old memory address
						RequestWrite <= 1'b1;
						RequestRead <= 1'b0;
						RequestWriteData <= ProcLine[31:0];
						State <= stWriteProc0;						
				end	else if ((~ProcTagMatch || ~ProcLineValid) && (dc_Read || dc_Write)) begin
						$display("%d: DCache: Invalid Line - Reading From Memory",$time);
						RequestAddress <=	{ProcReqTag, ProcLineNum,2'b00}; //set address to new memory address
						RequestRead <= 1'b1;
						RequestWrite <= 1'b0;
						State <= stReadProc0;										
						
				end else if ((ProcTagMatch && ProcLineValid) && dc_Write) begin
					//handle writing to cache
					if (dc_Size == BYTE) begin
						$display("%d: DCache: Writing Byte to Cache.  Addr: %h, Line: %0d, Bits: %0d-%0d, Data: %h",$time,dc_Address,ProcLineNum,8*(dc_Address[3:0]+1)-1,8*(dc_Address[3:0]+1)-8,dc_DataFromProcessor[7:0]);
						cacheLineData[ProcLineNum][ 8*(dc_Address[3:0]+1)-1 -: 8] <= dc_DataFromProcessor[7:0];
					end else if (dc_Size == HALF) begin
						$display("%d: DCache: Writing Half to Cache.  Addr: %h, Line: %0d, Bits: %0d-%0d, Data: %h",$time,dc_Address,ProcLineNum,16*(dc_Address[3:1]+1)-1,16*(dc_Address[3:1]+1)-16,dc_DataFromProcessor[15:0]);
						cacheLineData[ProcLineNum][16*(dc_Address[3:1]+1)-1 -: 16] <= dc_DataFromProcessor[15:0];
					end else if (dc_Size == WORD) begin																																	   
						$display("%d: DCache: Writing Word to Cache.  Addr: %h, Line: %0d, Bits: %0d-%0d, Data: %h",$time,dc_Address,ProcLineNum,32*(dc_Address[3:2]+1)-1,32*(dc_Address[3:2]+1)-32,dc_DataFromProcessor[31:0]);
						cacheLineData[ProcLineNum][32*(dc_Address[3:2]+1)-1 -: 32] <= dc_DataFromProcessor[31:0];
					end	
					cacheLineStatusBits[ProcLineNum][DirtyBit] <= 1'b1;
				end
			end		   
			
			
			stWriteProc0: begin
				if (dc_RequestReady) begin
					RequestAddress <=  {ProcTag, ProcLineNum,2'b01};
					RequestWriteData <= ProcLine[63:32];
					RequestWrite <= 1'b1;
					State <= stWriteProc1;
				end else
					RequestWrite <= 1'b0;
			end
			
			stWriteProc1: begin
				if (dc_RequestReady) begin
					RequestAddress <=  {ProcTag, ProcLineNum,2'b10};
					RequestWriteData <= ProcLine[95:64];
					RequestWrite <= 1'b1;
					State <= stWriteProc2;
				end else
					RequestWrite <= 1'b0;
			end
			
			stWriteProc2: begin
				if (dc_RequestReady) begin
					RequestAddress <=  {ProcTag, ProcLineNum,2'b11};
					RequestWriteData <= ProcLine[127:96];
					RequestWrite <= 1'b1;
					State <= stWriteProc3;
				end else
					RequestWrite <= 1'b0;
			end
			
			stWriteProc3: begin
				RequestWrite <= 1'b0;
				if (dc_RequestReady) begin								  
					//just clear out the dirty bit
					cacheLineStatusBits[ProcLineNum][DirtyBit] <= 1'b0;   
					State <= stIDLE;
				end					 
			end	 
			
			stReadProc0: begin
				if (dc_RequestReady) begin
					RequestAddress <=	{ProcReqTag, ProcLineNum,2'b01};
					cacheLineData[ProcLineNum][31:0] <= dc_DataFromMem;
					RequestRead <= 1'b1;
					State <= stReadProc1;
				end else
					RequestRead <= 1'b0;
			end		
			
			stReadProc1: begin
				if (dc_RequestReady) begin
					RequestAddress <=	{ProcReqTag, ProcLineNum,2'b10};
					cacheLineData[ProcLineNum][63:32] <= dc_DataFromMem;
					RequestRead <= 1'b1;
					State <= stReadProc2;
				end else
					RequestRead <= 1'b0;
			end
			
			stReadProc2: begin
				if (dc_RequestReady) begin
					RequestAddress <=	{ProcReqTag, ProcLineNum,2'b11};
					cacheLineData[ProcLineNum][95:64] <= dc_DataFromMem;
					RequestRead <= 1'b1;
					State <= stReadProc3;
				end else
					RequestRead <= 1'b0; 
			end		
			
			stReadProc3: begin
				RequestRead <= 1'b0;
				if (dc_RequestReady) begin
					cacheLineData[ProcLineNum][127:96] <= dc_DataFromMem;
					cacheLineTag[ProcLineNum] <= ProcReqTag;
					cacheLineStatusBits[ProcLineNum][DirtyBit] <= 1'b0;
					cacheLineStatusBits[ProcLineNum][ValidBit] <= 1'b1;	   
					State <= stIDLE;
				end
			end
			
			default: begin
				RequestRead <= 1'b0;
				RequestWrite <= 1'b0;
				State <= stIDLE;
			end
		endcase	
		end
	end
	
	//synthesis translate_off
	integer i;
	initial begin
		for (i=0; i<NUMLINES; i=i+1) begin
			cacheLineData[i] <= 0;
			cacheLineTag[i] <= 0;
			cacheLineStatusBits[i] <= 0;   
		end
	end	   
	//synthesis translate_on
	
endmodule