`timescale 1ns/1ps

module InstructionCache(
	input 			clk,
	input 			reset,
	
	//Connections to CoherencyEngine
	input	[31:0]	ic_DataFromMem,
	output	[29:0]	ic_RequestAddr,
	output			ic_RequestRead,
	input			ic_RequestReady,	
	
	//Connections to Processor
	output	[31:0]	ic_DataToProcessor,
	input	[31:0]	ic_Address,
	output			ic_Stall	
	);
	
	localparam stIDLE 				=0;
	
	localparam stReadProc0			=1;
	localparam stReadProc1			=2;
	localparam stReadProc2			=3;
	localparam stReadProc3			=4;
	
	
	//32 Bit Address:
	//Bits:     Use:
	//1-0		Byte Address
	//3-2		Word Address
	//(LINENUMBITS+3)-4		Line
	//31-(LINENUMBITS+4)	Tag
	
	//Parameter Definitions
	parameter NUMLINES = 16;
	parameter LINENUMBITS = 4;
	parameter TAGBITS = 32 - LINENUMBITS - 4;
	
	localparam ValidBit =0;
	localparam DirtyBit =1;
	
	//cacheLine declarations
	reg [127:0] cacheLineData [0:NUMLINES-1];
	reg [TAGBITS-1:0] cacheLineTag [0:NUMLINES-1];
	reg cacheLineStatusBits [0:NUMLINES-1];
	
	//wires to hold sections of addresses
	wire [LINENUMBITS-1:0] ProcLineNum, InvalidateLineNum;
	wire [TAGBITS-1:0] ProcReqTag, InvalidateReqTag;
	
	//get line number and tag bits from processor address
	assign ProcLineNum = ic_Address[(LINENUMBITS+3):4];
	assign ProcReqTag = ic_Address[31:LINENUMBITS+4];
	
	//wires containing line addressed by LINENUM portion of address
	wire [127:0] ProcLine;
	wire [TAGBITS-1:0] ProcTag;
	wire ProcLineValid, ProcLineDirty;
	wire ProcTagMatch;
	
	assign ProcLine = cacheLineData[ProcLineNum];
	assign ProcTag = cacheLineTag[ProcLineNum];
	assign ProcLineValid = cacheLineStatusBits[ProcLineNum];
	assign ProcTagMatch = ProcTag == ProcReqTag;
		
	assign ic_Stall = (!ProcLineValid || !ProcTagMatch);
	
	reg [29:0]  RequestAddress;	 
	reg			RequestRead;	
	reg [2:0] State;				
	//if the state machine is idle, use the current address from 
	//the processor as the request address to the memory controller
	assign ic_RequestAddr = (State == stIDLE) ? ic_Address[31:4] : RequestAddress;
	
	//Invalidation is done when InvalidateLineValid is false, or the tags do not match
						
	assign ic_RequestRead = RequestRead;
	
	//Assign output to the processor based on cache line, etc.	
	wire [31:0] selectedWord;
	wire [31:0] ic_DataToProcessor;
	
	assign selectedWord = 
	(ic_Address[3:2] == 2'b00) ? cacheLineData[ProcLineNum][31:0] :
	(ic_Address[3:2] == 2'b01) ? cacheLineData[ProcLineNum][63:32] :
	(ic_Address[3:2] == 2'b10) ? cacheLineData[ProcLineNum][95:64] :
								 cacheLineData[ProcLineNum][127:96];
		
	assign ic_DataToProcessor = selectedWord[31:0];
	
	
	
	
	always @(posedge clk) begin
		if(reset) begin
			State <= stIDLE;
			RequestRead <= 1'b0;
			RequestAddress <= 0;
			cacheLineStatusBits[0] <= 0;
			cacheLineStatusBits[1] <= 0;
			cacheLineStatusBits[2] <= 0; 
			cacheLineStatusBits[3] <= 0;
			cacheLineStatusBits[4] <= 0;
			cacheLineStatusBits[5] <= 0;
			cacheLineStatusBits[6] <= 0;
			cacheLineStatusBits[7] <= 0;
			cacheLineStatusBits[8] <= 0;
			cacheLineStatusBits[9] <= 0;
			cacheLineStatusBits[10] <= 0;
			cacheLineStatusBits[11] <= 0;
			cacheLineStatusBits[12] <= 0;
			cacheLineStatusBits[13] <= 0;
			cacheLineStatusBits[14] <= 0;
			cacheLineStatusBits[15] <= 0;
		end
		else begin
	   	case(State)
			stIDLE: begin						
				if ((~ProcTagMatch || ~ProcLineValid)) begin
						RequestAddress <=	{ProcReqTag, ProcLineNum,2'b00}; //set address to new memory address
						RequestRead <= 1'b1;
						State <= stReadProc0;										
				end
			end		   
					
			stReadProc0: begin
				if (ic_RequestReady) begin
					cacheLineData[ProcLineNum][31:0] <= ic_DataFromMem;
					RequestAddress <=	{ProcReqTag, ProcLineNum,2'b01};
					RequestRead <= 1'b1;
					State <= stReadProc1;
				end else
					RequestRead <= 1'b0;
			end		
			
			stReadProc1: begin
				if (ic_RequestReady) begin
					cacheLineData[ProcLineNum][63:32] <= ic_DataFromMem;
					RequestAddress <=	{ProcReqTag, ProcLineNum,2'b10};
					RequestRead <= 1'b1;
					State <= stReadProc2;
				end else
					RequestRead <= 1'b0;
			end
			
			stReadProc2: begin
				if (ic_RequestReady) begin
					cacheLineData[ProcLineNum][95:64] <= ic_DataFromMem;
					RequestAddress <=	{ProcReqTag, ProcLineNum,2'b11};
					RequestRead <= 1'b1;
					State <= stReadProc3;
				end else
					RequestRead <= 1'b0; 
			end		
			
			stReadProc3: begin
				RequestRead <= 1'b0;
				if (ic_RequestReady) begin
					cacheLineData[ProcLineNum][127:96] <= ic_DataFromMem;
					cacheLineTag[ProcLineNum] <= ProcReqTag;
					cacheLineStatusBits[ProcLineNum] <= 1'b1;	   
					State <= stIDLE;
				end
			end
			
			default: begin
				RequestRead <= 1'b0;
				State <= stIDLE;
			end
		endcase	
		end
	end
	
	
	integer i;
	initial begin
		for (i=0; i<NUMLINES; i=i+1) begin
			cacheLineData[i] <= 0;
			cacheLineTag[i] <= 0;
			cacheLineStatusBits[i] <= 0;   
		end
			
	end
	
endmodule