`timescale 1ns/1ps

/*
IO Arbiter
-----------------------------------------------
Revision 1 - Adrian Caulfield - April 29, 2007
Fixed bug where write signal asserted for more than one
	cycle caused incorrect stalling
-----------------------------------------------
Revision 0 - Adrian Caulfield - April 23, 2007
*/

module IOArbiter
	(
	input clk,
	input reset,
	
	output				p1_Stall,
	output		[31:0]	p1_DataOut,
	input		[31:0]	p1_DataIn,
	input				p1_Read,
	input				p1_Write,
	input		[1:0]	p1_Size,
	input		[31:0]	p1_Address,
	
	output				p2_Stall,
	output		[31:0]	p2_DataOut,
	input		[31:0]	p2_DataIn,
	input				p2_Read,
	input				p2_Write,
	input		[1:0]	p2_Size,
	input		[31:0]	p2_Address,
	
	output wire	[31:0] 	io_DataOut,
	output wire	[1:0]	io_Size,
	output wire	[31:0] 	io_Address,
	
	
	output reg			io1_WE,
	output reg			io2_WE,
	output reg			io3_WE,
	output reg			io4_WE,
	output reg			io5_WE,
	output reg			io6_WE,
	
	output reg			vgaChar_WE,
	output reg			vgaColor_WE,
	
	input		[31:0]	io1_DataIn,
	input		[31:0]	io2_DataIn,
	input		[31:0]	io3_DataIn,
	input		[31:0]	io4_DataIn,
	input		[31:0]	io5_DataIn,
	input		[31:0]	io6_DataIn
	);
	
	reg		[31:0]	io_DataIn;		  
	reg				io_Ready;
	wire			io_Read, io_Write;	  
	reg 			p1_Pending, p2_Pending;	   
	wire			p1_Ready, p2_Ready;
	
	assign p1_Stall = ((p1_Read || p1_Write)&&~p1_Pending) || (p1_Pending && ~p1_Ready);
	assign p2_Stall = ((p2_Read || p2_Write)&&~p2_Pending) || (p2_Pending && ~p2_Ready);
		

	
	MemoryArbiter #(.AddressWidth(32)) memArb(
	.clk(clk),
	.reset(reset),
	
	.p1_Read(p1_Read && ~p1_Pending),
	.p1_Write(p1_Write && ~p1_Pending),	
	.p1_Size(p1_Size),
	.p1_DataToMem(p1_DataIn),
	.p1_Address(p1_Address),
	.p1_DataFromMem(p1_DataOut),
	.p1_Ready(p1_Ready),
	
	.p2_Read(p2_Read && ~p2_Pending),
	.p2_Write(p2_Write && ~p2_Pending),
	.p2_Size(p2_Size),
	.p2_DataToMem(p2_DataIn),
	.p2_Address(p2_Address),
	.p2_DataFromMem(p2_DataOut),
	.p2_Ready(p2_Ready),
	
	.mem_Read(io_Read),
	.mem_Write(io_Write),
	.mem_Size(io_Size),
	.mem_DataToMem(io_DataOut),
	.mem_Address(io_Address),
	.mem_DataFromMem(io_DataIn),
	.mem_Ready(io_Ready)
	);
	
	
	always @(posedge clk) begin
		if(reset) begin
			p1_Pending <= 0;
			p2_Pending <= 0;
		end
		
		else begin
		//io Always Completes in one Cycle
		io_Ready <= io_Read || io_Write;
		
		//track whether we need to stall...
		p1_Pending <= ((p1_Read || p1_Write) && ~p1_Ready) || (p1_Pending && ~p1_Ready);
		p2_Pending <= ((p2_Read || p2_Write) && ~p2_Ready) || (p2_Pending && ~p2_Ready);
		end
		
	end
	
		
	//IO Selection
	always @(*) begin	
		io1_WE	<= 1'b0; 
		io2_WE	<= 1'b0;
		io3_WE	<= 1'b0;
		io4_WE	<= 1'b0;
		io5_WE	<= 1'b0;
		io6_WE	<= 1'b0;
		vgaChar_WE <= 1'b0;
		vgaColor_WE <= 1'b0;
		case(io_Address[15:13])
			3'b110: begin
				io_DataIn	<= 32'b0;
				vgaChar_WE	<= io_Write;
			end
			3'b111: begin
				io_DataIn	<= 32'b0;
				vgaColor_WE	<= io_Write;
			end
			default: begin
				case(io_Address[8:6])
					4'h0: begin
						io_DataIn 	<= io1_DataIn;
						io1_WE	  	<= io_Write;
					end
					4'h1: begin
						io_DataIn 	<= io2_DataIn;
						io2_WE		<= io_Write;
					end
					4'h2: begin
						io_DataIn	<= io3_DataIn;
						io3_WE		<= io_Write;
					end
					4'h3: begin
						io_DataIn	<= io4_DataIn;
						io4_WE		<= io_Write;
					end
					4'h4: begin
						io_DataIn	<= io5_DataIn;
						io5_WE		<= io_Write;
					end
					4'h5: begin
						io_DataIn	<= io6_DataIn;
						io6_WE		<= io_Write;
					end
					default: begin
						io_DataIn	<= 32'b0;
					end
				endcase
			end
		endcase
	end		
		
endmodule
