`timescale 1ns/1ps

/*
Memory System
--------------------
Adrian Caulfield (adrianmc) - April 23, 2007

*Two port memory controller - one read, one read/write.  
*All Memories are WORD addressable only
*Requests cannot be pipelined together, although a request on each port at the same time is valid
*Address must stay constant through cycle in which Ready signal is asserted

*Provides 32KB each for Instruction Memory, Data Memory, and Stack
*Location of Instruction, Data, and Stack are parameterizable with defaults set to:
 IRAM: 0x0040xxxx
 DRAM: 0x1000xxxx
STACK: 0x7fffxxxx

*A 2KB ready only bios is also included, defaulting to 0x0000xxxx

*Memory Accesses take minimum of 3 cycles to complete, although parameter can be changed to increase access time
*/

module Memory(
	input				clk,
	input 				reset,
	
	output 	reg[31:0] 	iDataFromMem,
	output 				iDataReady,
	input 	[29:0] 		iAddress,
	input 				iRead,
	
	output 	reg[31:0] 	dDataFromMem,
	output 				dDataReady,
	input 	[31:0] 		dDataToMem,   
	input 	[29:0] 		dAddress,
	input 				dWrite,
	input 				dRead
	);
	
	parameter BIOS_ADDR = 16'h0000;
	parameter IRAM_ADDR = 16'h0040;
	parameter DRAM_ADDR = 16'h1000;
	parameter STACK_ADDR = 16'h7fff;   
	
	parameter ACCESS_DELAY = 3;
	
	
	localparam ACCESS_DELAY_INTERNAL = (ACCESS_DELAY < 3 ? 3 : ACCESS_DELAY);
	
	wire [31:0] biosDataA;
	wire [31:0] iramDataA;
	wire [31:0] dramDataA;
	wire [31:0] stackDataA;

	wire [31:0] iramDataB;
	wire [31:0] dramDataB;
	wire [31:0] stackDataB; 
	
	reg [ACCESS_DELAY_INTERNAL:0] iReady, dReady;  
	
	assign iDataReady = iReady[ACCESS_DELAY_INTERNAL];
	assign dDataReady = dReady[ACCESS_DELAY_INTERNAL];
	
	
	always @(posedge clk) begin
		if (iAddress[29:14] == IRAM_ADDR) begin
			iDataFromMem <= iramDataA;
			end
		else if (iAddress[29:14] == DRAM_ADDR) begin
			iDataFromMem <= dramDataA;
		end
		else if (iAddress[29:14] == STACK_ADDR) begin
			iDataFromMem <= stackDataA;	 	   
				end
		else if (iAddress[29:14] == BIOS_ADDR) begin
			iDataFromMem <= biosDataA;		 		
		end else begin
		iDataFromMem <= 32'b0;
		end
			
		if (dAddress[29:14] == IRAM_ADDR)
			dDataFromMem <= iramDataB;
		else if (dAddress[29:14] == DRAM_ADDR)
			dDataFromMem <= dramDataB;
		else if (dAddress[29:14] == STACK_ADDR)
			dDataFromMem <= stackDataB;		 
		else
			dDataFromMem <= 32'b0;
		
		/*Delay ready signal until data has propagated through the muxes*/
		iReady <= {iReady[ACCESS_DELAY_INTERNAL-1:0],iRead};
		dReady <= {dReady[ACCESS_DELAY_INTERNAL-1:0],dRead || dWrite};
		
		if (reset) begin
			iReady<=0;
			dReady<=0;	
		end
	end
	
	// 32KB per block 
	bram16 	iram(clk,iramDataA,iramDataB,iAddress[12:0],dAddress[12:0],dDataToMem,dWrite && dAddress[29:14] ==IRAM_ADDR);
	bram16 	dram(clk,dramDataA,dramDataB,iAddress[12:0],dAddress[12:0],dDataToMem,dWrite && dAddress[29:14] ==DRAM_ADDR);
	bram16 	stack(clk,stackDataA,stackDataB,iAddress[12:0],dAddress[12:0],dDataToMem,dWrite && dAddress[29:14] ==STACK_ADDR);
		
	
	/*// 8KB per block
	bram4 	iram(clk,iramDataA,iramDataB,iAddress[10:0],dAddress[10:0],dDataToMem,dWrite && dAddress[29:14] ==IRAM_ADDR);
	bram4 	dram(clk,dramDataA,dramDataB,iAddress[10:0],dAddress[10:0],dDataToMem,dWrite && dAddress[29:14] ==DRAM_ADDR);
	bram4 	stack(clk,stackDataA,stackDataB,iAddress[10:0],dAddress[10:0],dDataToMem,dWrite && dAddress[29:14] ==STACK_ADDR);
	*/
	
	lab4_rom lab4rom(clk,iAddress[9:0],biosDataA);
	
	
	//synthesis translate_off
	always @(posedge clk) begin
		if (dWrite) $display("%d: Memory: Write to %h, Data: %h",$time,{dAddress,2'b0},dDataToMem);
		if (dRead) $display("%d: Memory: Read from %h, Data: %h", $time, {dAddress, 2'b0}, dDataFromMem);
	end
	//synthesis translate_on

	
endmodule

//32KB Dual Port RAM, word addressable
module bram16(
	input	clk,
	output	reg [31:0] DA,
	output	reg [31:0] DB,
	input	[12:0] AA,
	input 	[12:0] AB,
	input	[31:0] DI,
	input	WE);
	
	wire [31:0] DA0,DA1,DA2,DA3,DB0,DB1,DB2,DB3;
	
	always @(posedge clk) begin
		case (AA[10:9])
			0: DA <= DA0;
			1: DA <= DA1;
			2: DA <= DA2;
			3: DA <= DA3;
		endcase
		
		case (AB[10:9])
			0: DB <= DB0;
			1: DB <= DB1;
			2: DB <= DB2;
			3: DB <= DB3;
		endcase
	end

	bram4 ram0(clk,DA0,DB0,AA[10:0],AB[10:0],DI,WE && AA[12:11]==0);
	bram4 ram1(clk,DA1,DB1,AA[10:0],AB[10:0],DI,WE && AA[12:11]==1);
	bram4 ram2(clk,DA2,DB2,AA[10:0],AB[10:0],DI,WE && AA[12:11]==2);
	bram4 ram3(clk,DA3,DB3,AA[10:0],AB[10:0],DI,WE && AA[12:11]==3);
	
	
endmodule
	
//8KB Dual Port RAM, word addressable
module bram4(	
	input 	clk,
	output 	reg [31:0] DA,
	output 	reg [31:0] DB,
	input	[10:0] AA,
	input	[10:0] AB,
	input	[31:0] DI,
	input	WE);
	
	
	wire [31:0] DA0,DA1,DA2,DA3,DB0,DB1,DB2,DB3;
	
	always @(posedge clk) begin
		case (AA[10:9])
			0: DA <= DA0;
			1: DA <= DA1;
			2: DA <= DA2;
			3: DA <= DA3;
		endcase
		
		case (AB[10:9])
			0: DB <= DB0;
			1: DB <= DB1;
			2: DB <= DB2;
			3: DB <= DB3;
		endcase
	end
	
	bram1 ram0(clk,DA0,DB0,AA[8:0],AB[8:0],DI,WE && AB[10:9]==0);
	bram1 ram1(clk,DA1,DB1,AA[8:0],AB[8:0],DI,WE && AB[10:9]==1);
	bram1 ram2(clk,DA2,DB2,AA[8:0],AB[8:0],DI,WE && AB[10:9]==2);
	bram1 ram3(clk,DA3,DB3,AA[8:0],AB[8:0],DI,WE && AB[10:9]==3);

	
endmodule


//2KB Dual Port RAM, word addressable
module bram1(
	input 	clk,
	output 	wire [31:0] DA,
	output 	wire [31:0] DB,
	input	[8:0] AA,
	input	[8:0] AB,
	input	[31:0] DI,
	input	WE);
	
	
	//synthesis translate_off
	`library("ram","virtex2p")
	//synthesis translate_on
	RAMB16_S36_S36 ram 
	(.DOA(DA),
	.DOB(DB),
	.DOPA(),
	.DOPB(),
	.ADDRA(AA),
	.ADDRB(AB),
	.CLKA(clk),
	.CLKB(clk),
	.DIA(32'b0),
	.DIB(DI),
	.DIPA(4'b0),
	.DIPB(4'b0),
	.ENA(1'b1),
	.ENB(1'b1),
	.SSRA(1'b0),
	.SSRB(1'b0),
	.WEA(1'b0),
	.WEB(WE));	  	
endmodule

module lab4_rom
	(input 			clk,
	input 			[9:0] addr,
	output wire 	[31:0]  DA);
	
	reg [31:0] id;
	assign DA = id;

	always @(posedge clk) begin
		case ({4'b0,addr[9:0],2'b0})
			`include "Assembly/bios.rom"
			default: id <= 32'b0;
		endcase
	end

	
endmodule
