`timescale 1ns/1ps

/*
Memory Arbiter
----------------------------------------------------------------------
Revision 1 - Adrian Caulfield - April 29, 2007

Incorrect parameterization of p1r_Address and p2r_Address signals corrected
----------------------------------------------------------------------
Revision 0 - Adrian Caulfield - April 23, 2007

Will handle one read/write from each port at a time,
however, requests from each port may overlap or occur simultaneous.
----------------------------------------------------------------------
Paramaters:
AddressWidth:  specifies the width of the Address inputs and outputs
*/

module MemoryArbiter(
	clk,
	reset,
	
	p1_Read,
	p1_Write,
	p1_DataToMem,
	p1_Size,
	p1_Address,
	p1_DataFromMem,
	p1_Ready,
	
	p2_Read,
	p2_Write,
	p2_DataToMem,
	p2_Size,
	p2_Address,
	p2_DataFromMem,
	p2_Ready,
	
	mem_Read,
	mem_Write,
	mem_DataToMem,
	mem_Size,
	mem_Address,
	mem_DataFromMem,
	mem_Ready
	
	);

	parameter AddressWidth = 30; 
	
	input				clk;
	input				reset;
	
	input				p1_Read;
	input				p1_Write;
	input	[31:0]		p1_DataToMem;
	input	[1:0]		p1_Size;
	input [AddressWidth-1:0] p1_Address;
	output	[31:0]		p1_DataFromMem;
	output				p1_Ready;
	
	
	input				p2_Read;
	input				p2_Write;
	input	[31:0]		p2_DataToMem;
	input	[1:0]		p2_Size;
	input [AddressWidth-1:0] p2_Address;
	output	[31:0]		p2_DataFromMem;
	output				p2_Ready;
	
	output	reg			mem_Read;
	output	reg			mem_Write;
	output	reg[31:0]	mem_DataToMem;
	output	reg[1:0]	mem_Size;
	output reg [AddressWidth-1:0] mem_Address;
	input	[31:0]		mem_DataFromMem;
	input				mem_Ready;
	
	
	/* Net Declarations */	
	wire p1_Act, p2_Act;
	wire ActBoth;
	reg PriorityBit;
	
	reg p1r_Read, p1r_Write;
	reg	[1:0]  p1r_Size;
	reg [31:0] p1r_DataToMem;
	reg [AddressWidth-1:0] p1r_Address;
	
	reg p2r_Read, p2r_Write;
	reg [1:0]  p2r_Size;
	reg [31:0] p2r_DataToMem;
	reg [AddressWidth-1:0] p2r_Address;
	
	reg [2:0] p1_State, p2_State;

	//State machine states for each port
	localparam IDLE=0, PENDING=1, ACTIVE=2;
	
	
	/* Continuous Assignments */
	assign p1_DataFromMem = mem_DataFromMem;
	assign p2_DataFromMem = mem_DataFromMem;
	
	
	assign p1_Act = p1_Read || p1_Write;
	assign p2_Act = p2_Read || p2_Write;
	assign ActBoth = p1_Act && p2_Act;
	
	
	//assign ready signals
	assign p1_Ready = p1_State == ACTIVE && mem_Ready;
	assign p2_Ready = p2_State == ACTIVE && mem_Ready;
	
	
	
	
	always @(posedge clk) begin
		if (reset) begin
			p1_State <= IDLE;
			p2_State <= IDLE;
			PriorityBit <= 1'b0;
		end	else begin
		
		//toggle priority bit
		PriorityBit <= ~PriorityBit;
		
		//default to 0
		mem_Read <= 1'b0;
		mem_Write <= 1'b0;
		
		
		//port 1 activity
		if (p1_Act) begin
			if ((p2_State != IDLE && ~p2_Ready) || ((p2_State == IDLE || p2_Ready) && (p2_Act && PriorityBit))) begin
				//already handling a second request, or two simulataneous requests and p2 gets priority
				p1r_Read <= p1_Read;
				p1r_Write <= p1_Write;
				p1r_Size <= p1_Size;
				p1r_DataToMem <= p1_DataToMem;
				p1r_Address <= p1_Address;
				p1_State <= PENDING;
			end else begin
				mem_Read <= p1_Read;
				mem_Write <= p1_Write;
				mem_Size <= p1_Size;
				mem_DataToMem <= p1_DataToMem;
				mem_Address <= p1_Address;
				p1_State <= ACTIVE;
			end					   
		end
		
		//port 2 activity
		if (p2_Act) begin
			if ((p1_State != IDLE && ~p1_Ready) || ((p1_State == IDLE || p1_Ready) && (p1_Act && ~PriorityBit))) begin
				p2r_Read <= p2_Read;
				p2r_Write <= p2_Write; 
				p2r_Size <= p2_Size;
				p2r_DataToMem <= p2_DataToMem;
				p2r_Address <= p2_Address;
				p2_State <= PENDING;
			end else begin			
				mem_Read <= p2_Read;
				mem_Write <= p2_Write;
				mem_Size <= p2_Size;
				mem_DataToMem <= p2_DataToMem;
				mem_Address <= p2_Address;
				p2_State <= ACTIVE;
			end
		end
		
		//reset Ready signals to zero every cycle
		
		if (p1_Ready) begin
			p1_State <= IDLE;
			if (p2_State == PENDING) begin
				mem_Read <= p2r_Read;
				mem_Write <= p2r_Write;
				mem_Size <= p2r_Size;
				mem_DataToMem <= p2r_DataToMem;
				mem_Address <= p2r_Address;
				p2_State <= ACTIVE;
			end
		end else if (p2_Ready) begin
			p2_State <= IDLE;	
			if (p1_State == PENDING) begin
				mem_Read <= p1r_Read;
				mem_Write <= p1r_Write;
				mem_Size <= p1r_Size;
				mem_DataToMem <= p1r_DataToMem;
				mem_Address <= p1r_Address;
				p1_State <= ACTIVE;
			end
		end

		if (p1_State == IDLE && p2_State == PENDING) begin
			mem_Read <= p2r_Read;
			mem_Write <= p2r_Write;	 
			mem_Size <= p2r_Size;
			mem_DataToMem <= p2r_DataToMem;
			mem_Address <= p2r_Address;
			p2_State <= ACTIVE;
		end					   
		
		if (p2_State == IDLE && p1_State == PENDING) begin
			mem_Read <= p1r_Read;
			mem_Write <= p1r_Write;
			mem_Size <= p1r_Size;
			mem_DataToMem <= p1r_DataToMem;
			mem_Address <= p1r_Address;
			p1_State <= ACTIVE;
		end
			
		//shouldn't really happen... but just incase:
		if (p1_State == PENDING && p2_State == PENDING) begin
			if (PriorityBit) begin
				mem_Read <= p2r_Read;
				mem_Write <= p2r_Write;
				mem_Size <= p2r_Size;
				mem_DataToMem <= p2r_DataToMem;
				mem_Address <= p2r_Address;
				p2_State <= ACTIVE;
			end else begin		   
				mem_Read <= p1r_Read;
				mem_Write <= p1r_Write;
				mem_Size <= p1r_Size;
				mem_DataToMem <= p1r_DataToMem;
				mem_Address <= p1r_Address;
				p1_State <= ACTIVE;
			end
		end
		
		
		end
	end
endmodule






